% phaseCorrectData - perform EPI phase correctiion and VRGF resampling
%
% Marquette University,   Milwaukee, WI  USA
% Copyright 2010 - All rights reserved.
%
% Author: Fred Frigo    02-Nov-2010
%                       10-Nov-2016 - Updates for new Pfile header format.
%                                     Use "ref.h5" file instead of ref.dat
%                                     for phase correction coefficeints
%
% @param pfile the path=filename to Pfile for scan parameters
% @param pname the path to "ref.h5" and "vrgf.dat" files
% @param slice_no - slice number within pass
% @param num_channels - number of channels
% @param input_raw_data - data before VRGF re-sampling & phase correction
%
% @return an array containing phase corrected and VRGF resampled raw data 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



function raw_data = phaseCorrectData(pfile, pname, slice_no, num_channels, input_raw_data)
    i = sqrt(-1);
    my_slice = slice_no;
            
    % Open Pfile to read scan parameters.
    fid = fopen(pfile,'r', 'ieee-be');
    if fid == -1
        err_msg = sprintf('Unable to locate Pfile %s', pfile)
        return;
    end

    % Determine size of Pfile header based on Rev number
    status = fseek(fid, 0, 'bof');
    [f_hdr_value, count] = fread(fid, 1, 'real*4');
    rdbm_rev_num = f_hdr_value(1);
    if( rdbm_rev_num == 7.0 )
        pfile_header_size = 39984;  % LX
    elseif ( rdbm_rev_num == 8.0 )
        pfile_header_size = 60464;  % Cardiac / MGD
    elseif (( rdbm_rev_num > 5.0 ) && (rdbm_rev_num < 6.0)) 
        pfile_header_size = 39940;  % Signa 5.5
    else
        % In 11.0 (ME2) the header and data are stored as little-endian
        fclose(fid);
        fid = fopen(pfile,'r', 'ieee-le');
        status = fseek(fid, 0, 'bof');
        [rdbm_rev_num, count] = fread(fid, 1, 'real*4');
        if (rdbm_rev_num == 9.0)  % 11.0 product release
            pfile_header_size= 61464;
        elseif (rdbm_rev_num == 11.0)  % 12.0 product release
            pfile_header_size= 66072;
        elseif (rdbm_rev_num > 11.0) & (rdbm_rev_num < 26.0)  
            % For 14.0 to 25.0 Pfile header size can be found here
            status = fseek(fid, 1468, 'bof');
            pfile_header_size = fread(fid,1,'integer*4');   
            status = fseek(fid, 1508, 'bof');
            prescan_offset = fread(fid,1,'integer*4');
        elseif ( rdbm_rev_num >= 26.0) & (rdbm_rev_num < 100.0)  % 26.0 to ??
            % For 26.0 Pfile header size moved to location just after rev num
            status = fseek(fid, 4, 'bof');
            pfile_header_size = fread(fid,1,'integer*4');
            status = fseek(fid, 44, 'bof');
            prescan_offset = fread(fid,1,'integer*4');              
        else
            err_msg = sprintf('Invalid Pfile header revision: %f', f_hdr_value )
            return;
        end
    end

    
    % Read header information
    if (rdbm_rev_num < 26.0 )
        status = fseek(fid, 0, 'bof');
    else
        status = fseek(fid, 76, 'bof'); % skip 76 bytes of data added for 26.0
    end
 

    % Read header information
    [hdr_value, count] = fread(fid, 102, 'integer*2');
    
    rhtype = hdr_value(29);
    npasses = hdr_value(33);
    nslices = hdr_value(35);
    nechoes = hdr_value(36);
    nframes = hdr_value(38);
    point_size = hdr_value(42);
    da_xres = hdr_value(52);
    da_yres = hdr_value(53);
    rc_xres = hdr_value(54);
    rc_yres = hdr_value(55);
    start_recv = hdr_value(101);
    stop_recv = hdr_value(102);
    nreceivers = (stop_recv - start_recv) + 1;
    
    
    % For 24.0 and earlier - nreceivers is determined as follows
    if (rdbm_rev_num < 25.0)
        nreceivers = (stop_recv - start_recv) + 1;
    else % For 25.0 and later - compute nreceivers from prescan noise data
        nreceivers = 0;
        max_num_receivers = 128;
        coil_weight_offset = prescan_offset + 332;
        status = fseek(fid, coil_weight_offset, 'bof');
        [psc_receiver_weights, count] = fread(fid,max_num_receivers,'real*4');  % 128 Max
        fclose(fid);
        % Determine number of receivers for 25.0 and later from receiver weights
        noise_threshold = psc_receiver_weights(max_num_receivers) + 1.0;
        for psc_rec = 1:max_num_receivers
            if ( psc_receiver_weights(psc_rec) >= noise_threshold) 
                nreceivers = nreceivers + 1;
            end
        end
    end
    
    % Use number of rows to determin image size for EPI - FF
    rc_xres = da_yres-1;
    rc_yres = rc_xres;

    % Determine number of slices in this Pfile:  this does not work for all cases.
    slices_in_pass = nslices/npasses

    % Compute size (in bytes) of each frame, echo and slice
    data_elements = da_xres*2*(da_yres-1);
    frame_size = da_xres*2*point_size;
    echo_size = frame_size*da_yres;
    slice_size = echo_size*nechoes;
    mslice_size = slice_size*slices_in_pass;

    % Name of vrgf.dat file
    vrgf_file = strcat(pname, 'vrgf.dat');

    % For DV26 and later, the "ref.dat" file is replaced by "ref.h5"
    h5_format = 1; % Use 'ref.h5' for EPI phase correction coeffs.
    
    if( h5_format == 1)
        ref_h5_file = strcat(pname, 'ref.h5');
        h5_fid = H5F.open(ref_h5_file); 
        %
        % Need to add code for multiple slices
        % This will only read a single slice
        for r=1:nreceivers
            constant_coeff_string = sprintf('slice_0/channel_%d/ConstantCoefficients',r-1);
            cset_id = H5D.open(h5_fid,constant_coeff_string);
            epi_pc_con(:,r) = H5D.read(cset_id);
            H5D.close(cset_id);
            
            linear_coeff_string = sprintf('slice_0/channel_%d/LinearCoefficients',r-1);
            lset_id = H5D.open(h5_fid,linear_coeff_string);
            epi_pc_lin(:,r) = H5D.read(lset_id);
            H5D.close(lset_id);    
        end     
        H5F.close(h5_fid);
    % Use 'ref.dat' for EPI phase correction coefficients (DV25 and earlier)   
    else
        % Name of ref.dat file
        ref_file = strcat(pname, 'ref.dat');
    
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Open ref.dat file to read EPI phase correction coeffs     %
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        ref_fid = fopen(ref_file,'r', 'ieee-le');
        if ref_fid == -1
            err_msg = sprintf('Unable to locate ref.dat %s', pfile)
            return;
        end

        % Read EPI Phase correction coefficients (from ref.dat file)
        ref_offset = (slice_no-1)*1024*nreceivers;
        status = fseek(fid, ref_offset, 'bof');
    
        for r=1:nreceivers
            [epi_pc_con(:,r), count] = fread(fid, 512, 'real*4'); % Constant coefficients
            [epi_pc_lin(:,r), count] = fread(fid, 512, 'real*4'); % Linear coefficients
        end
        fclose(ref_fid);
    end
 
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Perform EPI phase correction before VRGF                 %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
     % create alteration vector 
     vector_size = da_xres;
     alt= [1:vector_size];
     vtmp = [1:vector_size]; 
     alt = 1.0;
     for m=1:da_xres
         if( mod(m,2) == 0 )
            alt(m)=1.0;
         else
            alt(m)=-1.0;
         end
     end

     for m=1:da_xres
        x_index(m) = -(da_xres/2)-1+m;
     end
        
     for r=1:nreceivers
        % Create zero filled K space matrix
        raw_frames = zeros( da_yres -1,da_xres);
        
        % DV27 swaps rows and columns in the PFile?
        pc_raw_data = input_raw_data(:,:,r)';
        
        % Read raw data for each row
        for m = 1:(da_yres-1)
           
            % apply alteration vector for each frame of data 
            vtmp=pc_raw_data(m,:).*alt;
                     
            % Fourier transform each row of raw data
            raw_frames(m,:)=ifft(vtmp, da_xres);
            % Compute phase correction vectors from linear+constant terms
            phase_temp=(x_index.*epi_pc_lin(m,r)) + epi_pc_con(m,r);
            phase_correct=cos(phase_temp) + sin(phase_temp)*i;
      
            % Multiply Fourier transformed data by phase correction vector
            raw_frames(m,:)=raw_frames(m,:).*phase_correct;
   
            % Inverse Fourier transform phase corrected data
            raw_frames(m,:)=fft(raw_frames(m,:), da_xres);
            % Apply alteration vector
            raw_frames(m,:)=raw_frames(m,:).*alt;
            
            % Debug plot
            %if( m == ((da_yres-1)/2))
            %    figure();
            %    subplot( 2,1, 1);
            %    plot(x_index, unwrap(angle(pc_raw_data(m,:))));
            %    title('Unwrapped phase at center of K-space');
            %    ylabel('Before Phase correction');
            %    subplot( 2,1, 2);
            %    plot(x_index, unwrap(angle(raw_frames(m,:))));
            %    ylabel('After Phase correction');
            %end
            
        end  % epi phase correction loop
        
        %Perform VRGF Re-sampling.
        raw_data(:,:,r) = do_vrgf(raw_frames(:,:), da_xres, rc_xres, da_yres-1, vrgf_file);
        
     end  % receiver loop

end